/*
 * Jean II '99
 *
 * E-Squirt command line program for Linux-IrDA...
 * Send : Open a stream socket and send the URL...
 * Recv : Open a stream socket and wait for URLs...
 *
 * The current single thread model work only with things that can be spawned
 * in the background. For lynx and audio, see audio_squirt.c
 *
 * Note : The string may be NULL terminated to make things safer.
 *
 * This program is released under the GPL license.
 */

#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>		/* Not needed */

#include <linux/types.h>
#include <linux/irda.h>

#ifndef AF_IRDA
#define AF_IRDA 23
#endif /* AF_IRDA */

#undef INCLUDE_NULL
/*#define BROWSER		"lynx %s"*/
#define BROWSER		"netscape %s &"

/*
 * This is the Ir service name (in IrIAS) that we are using
 */
#define BEACON_SERVICE_NAME	"hp:beacon"
#define SQUIRT_SERVICE_NAME	"hp:esquirt"

/* Memory allocation for discovery */
#define DISC_MAX_DEVICES 10
#define DISC_BUF_LEN	sizeof(struct irda_device_list) + \
			sizeof(struct irda_device_info) * DISC_MAX_DEVICES

/****************************** SENDER ******************************/

/* ---------------------------------------------------------------- */
/*
 * Discover the server address...
 */
int
discover(int	fd)
{
  struct irda_device_list *	list;		/* List of device */
  unsigned char		buf[DISC_BUF_LEN];	/* Actual memory allocation */
  int			daddr = 0x0;
  int			len;
#if 0
  int			i;
#endif

  /* Set the list to point to the correct place */
  list = (struct irda_device_list *) buf;
  len = DISC_BUF_LEN;

  /* Ask for the discovery log */
  if(getsockopt(fd, SOL_IRLMP, IRLMP_ENUMDEVICES, buf, &len))
    {
      perror("getsockopt-enum");
      return(-1);
    }

  /* Check it out ! */
  if(len > 0)
    {
#if 0
      /* Parse it quickly */
      for(i=0; i < list->len; i++)
	{
	  if(list->dev[i].hints[0] & HINT_COMPUTER)
	    {
	      daddr = list->dev[i].daddr;
	      break;
	    }
	}

      /* Check if we got something */
      if(daddr == 0x0)
	{
	  printf("Could not find any suitable device...\n");
	  return(-1);
	}
#else
      daddr = list->dev[0].daddr;
#endif
    }
  else
    {
      /* Discovery log empty */
      printf("No device in sight !\n");
      return(-1);
    }

  /* Return the address */
  return(daddr);
}

/* ---------------------------------------------------------------- */
/*
 * Send a squirt to the server...
 */
int
send_squirt(char *	string)
{
  struct sockaddr_irda	peer;		/* Client connected to us */
  int			fd;		/* The listening socket */
  unsigned char		buf[4098];	/* Send buffer */
  unsigned char		bufin[1024];	/* stdin buffer */
  int			daddr;		/* Destination address */
  int			actual;

  /* Do we have some data to send ? */
  if(string == (char *) NULL)
    {
      printf("Type the url to squirt :");
      fflush(stdout);
      if(fgets(bufin, sizeof(bufin), stdin) == NULL)
	return(-1);
      string = bufin;
    }

  /* Set the url in the data to send */
  sprintf(buf, "%s", string);
  printf("Sending ``%s'' on service ``%s''\n",
	 buf, SQUIRT_SERVICE_NAME);

  /* Create socket - Stream/Sequenced packet */
  fd = socket(AF_IRDA, SOCK_STREAM, 0);
  /*fd = socket(AF_IRDA, SOCK_SEQPACKET, 0);*/
  if(fd < 0)
    {
      perror("Creating socket");
      return(-1);
    }

  /* Discover an address */
  daddr = discover(fd);
  if(daddr == -1)
    return(-1);

  /* Try to connect */
  peer.sir_family = AF_IRDA;
  strncpy(peer.sir_name, SQUIRT_SERVICE_NAME, 25);
  peer.sir_lsap_sel = LSAP_ANY;
  peer.sir_addr = daddr;
  if(connect(fd, (struct sockaddr*) &peer, sizeof(struct sockaddr_irda)))
    {
      perror("connect");
      return -1;
    }

#ifdef INCLUDE_NULL
  /* Send the data including the trailing '\0' */
  actual = write(fd, buf, strlen(buf) + 1);
  if(actual != (strlen(buf) + 1))
     {
       perror("write");
       return(-1);
     }
#else
  /* Send the data excluding the trailing '\0' */
  actual = write(fd, buf, strlen(buf));
  if(actual != (strlen(buf)))
     {
       perror("write");
       return(-1);
     }
#endif

  /* The end */
  close(fd);
  printf("Done...\n");
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Create XML string
 */
int
create_xml(char *	buf,
	   int		num,
	   char *	args[])
{
  /* Check number of args */
  switch(num)
    {
    case 0:
      strcpy(buf, "");
      break;

    case 1:
      strcpy(buf, args[0]);
      break;

    case 2:
      sprintf(buf, "<squirt label=\"%s\" url=\"%s\"\\>", args[0], args[1]);
      break;	      

    default:
      sprintf(buf, "<squirt label=\"%s\" url=\"%s\" type=\"%s\"\\>",
	      args[0], args[1], args[2]);
      break;
    }

  return 0;
}

/***************************** RECEIVER *****************************/

/* ---------------------------------------------------------------- */
/*
 * Parse an XML token
 */
char *
parse_token(char *	string,
	    char *	id,
	    char *	result)
{
  char *		pos;
  char *		pos2;
  char *		pos3;

  /* Find the ID */
  pos = strstr(string, id);
  if(pos == NULL)
    return(NULL);

  /* Find the first `"' */
  pos2 = strchr(pos, '"');
  if(pos2 == NULL)
    return(NULL);
  pos2++;

  /* Find the second `"' */
  pos3 = strchr(pos2, '"');
  if(pos3 == NULL)
    return(NULL);

  /* Copy over */
  memcpy(result, pos2, pos3 - pos2);
  result[pos3 - pos2] = '\0';
  return(result);
}

/* ---------------------------------------------------------------- */
/*
 * Parse an XML string
 */
int
parse_xml(char *	string,
	  char *	url)
{
  unsigned char		field[383];
  char *		token;

  /* Check if XML or not */
  if(string[0] != '<')
    {
      printf("  The URL is ``%s''\n", string);
      strcpy(url, string);
    }
  else
    {
      /* Find the label */
      token = parse_token(string, "label=", field);
      if(token != NULL)
	printf("  The Label is ``%s''\n", token);

      /* Find the label */
      token = parse_token(string, "url=", field);
      if(token != NULL)
	{
	  printf("  The XML URL is ``%s''\n", token);
	  strcpy(url, token);
	}

      /* Find the label */
      token = parse_token(string, "type=", field);
      if(token != NULL)
	printf("  The Type is ``%s''\n", token);
    }
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * After the main server receive an incoming connection, it fork
 * this function to handle the client...
 */
int
recv_squirt(int	conn_fd)
{
  unsigned char		buf[1024];	/* Receive buffer */
  int			ret;
  int			len;
  unsigned char		url[1024];

  printf("Incoming connection !\n");

  /* Read all packets that the client want to send */
  do
    {
      /* Read one packet */
      ret = recv(conn_fd, &buf, sizeof(buf), 0);
      if(ret <= 0)
	break;

      /* Display it */
      len = ret;
      buf[len] = '\0';	/* Just to avoid crash on incorrect data */
      printf("Received E-Squirt ``%s''\n", buf);
    }
  while(ret > 0);

  /* Check that we got something */
  if(len > 0)
    {
      /* Do something about the squirt... */
      strcpy(url, "");
      parse_xml(buf, url);
#ifdef BROWSER
      sprintf(buf, BROWSER, url);
      system(buf);
#endif
    }

  /* Cleanup */
  close(conn_fd);
  printf("Client disconnected...\n");
  return(0);
}

/****************************** SERVER ******************************/

/* ---------------------------------------------------------------- */
/*
 * Main server task : listen on the socket and fork process to
 * handle incomming connections...
 */
int
prepare_socket(char *	service)
{
  int			fd;	/* The socket */
  struct sockaddr_irda	self;		/* Socket address */
  int			addrlen;

  /* Create socket - Stream/Sequenced packet */
  fd = socket(AF_IRDA, SOCK_STREAM, 0);
  /*fd = socket(AF_IRDA, SOCK_SEQPACKET, 0);*/
  if(fd < 0)
    {
      perror("Creating socket (no IrDA stack ?)");
      return(-1);
    }

  /* Bind to the specific service we want to receive */
  self.sir_family = AF_IRDA;
  strncpy(self.sir_name, service, 25);
  self.sir_lsap_sel = LSAP_ANY;
  if(bind(fd, (struct sockaddr*) &self, sizeof(struct sockaddr_irda)))
    {
      perror("Connect");
      return(-1);
    }

  /* Set the socket in listen mode (indicate that we want packets) */
  if(listen(fd, 8))
    {
      perror("Listen");
      return(-1);
    }

  /* Get out info (optional) */
  addrlen = sizeof(struct sockaddr_irda);
  if (getsockname(fd, (struct sockaddr*) &self, &addrlen)) {
    perror("getsockname");
    return -1;
  }

  printf("IrDA E-Squirt receiver ready, service ``%s'' on LSAP 0x%02X...\n",
	 service, self.sir_lsap_sel);

  return(fd);
}

/* ---------------------------------------------------------------- */
/*
 * Main server task : listen on the socket and fork process to
 * handle incomming connections...
 */
int
server(void)
{
  struct sockaddr_irda	peer;		/* Client connected to us */
  int			addrlen;
  int			squirt_fd;	/* The e-Squirt listening socket */
  int			beacon_fd;	/* The Beacon listening socket */
  int			current_fd;	/* The current socket */
  int			conn_fd;	/* The data socket */
  fd_set		rfds;		/* File descriptors for select */
  struct timeval	tv;		/* Select timeout */
  int			retval;

  /* Create sockets and put them in listen mode */
  squirt_fd = prepare_socket(SQUIRT_SERVICE_NAME);
  beacon_fd = prepare_socket(BEACON_SERVICE_NAME);

  /* Receive packets */
  while(1)
    {
      printf("Waiting for E-Squirt connections...\n");

      /* Set stdin and our socket to listen to... */
      FD_ZERO(&rfds);
      FD_SET(0, &rfds);		/* stdin */
      FD_SET(squirt_fd, &rfds);	/* Ir socket */
      FD_SET(beacon_fd, &rfds);	/* Ir socket */
      /* Wait up to sixty seconds */
      tv.tv_sec = 60;
      tv.tv_usec = 0;

      current_fd = (squirt_fd > beacon_fd) ? squirt_fd : beacon_fd;

      /* Let's try a select (optional) */
      retval = select(current_fd + 1, &rfds, NULL, NULL, NULL);
      /*retval = select(current_fd + 1, &rfds, NULL, NULL, &tv);*/

      /* Check if timeout */
      if(!retval)
	continue;

      /* Check if stdin */
      if(FD_ISSET(0, &rfds))
	{
	  printf("Key pressed...\n");
	  continue;
	  break;
	}

      /* Check if on of the socket is ready */
      current_fd = -1;
      if((FD_ISSET(squirt_fd, &rfds)))
	current_fd = squirt_fd;
      if((FD_ISSET(beacon_fd, &rfds)))
	current_fd = beacon_fd;

      /* Check if we have one socket */
      if(current_fd < 0)
	{
	  printf("Unknown interuption...\n");
	  continue;
	}

      /* Duplicate the socket for receiving data */
      addrlen = sizeof(struct sockaddr_irda);
      conn_fd = accept(current_fd, (struct sockaddr *) &peer, &addrlen);
      if(conn_fd < 0) {
	perror("accept");
	return(-1);
      }

      /* Process the data */
      recv_squirt(conn_fd);
      /* Note : will close conn_fd */

      /* Now, go back to listen... */
    }

  /* Time to clean-up everything */
  close(squirt_fd);
  close(beacon_fd);
  printf("Exiting !\n");
  return(0);
}

/******************************* MAIN *******************************/

/* ---------------------------------------------------------------- */
/*
 * main body of the program
 */
int
main(int	argc,
     char *	argv[])
{
  char			string[1024];	/* XML string */

  /* Look if there is any command line args */
  if(argc > 1)
    {
      create_xml(string, argc - 1, argv + 1);
      send_squirt(string);
    }
  else
    server();

  return(0);
}
