/*
 * Jean II '99
 *
 * Send beacon info or receive beacon info
 * Send : Open a Ultra socket and send one packet...
 * Recv : Open a Ultra socket and wait for packets...
 *
 * Note : Ultra can carry only 382 bytes / packet, and this including
 * the ultra header (the UPID). As we use a one byte UPID, that's 381 bytes.
 *
 * Note : The string may be NULL terminated to make things safer.
 *
 * This program is released under the GPL license.
 */

#include <sys/socket.h>
#include <sys/types.h>

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <linux/types.h>
#include <linux/irda.h>

#ifndef AF_IRDA
#define AF_IRDA 23
#endif /* AF_IRDA */

/* To send beacons with the old XML format */
#undef OLD_XML

/*
 * This is the Ultra protocol ID we are currently using (7 bits)
 * Currently, IrDA-Ultra defines only UPID 0x1 for Obex. Maybe
 * we should use a bigger number to avoid clashes with future
 * standardisations...
 */
#define UPID_BEACON_TEMP	0x2		/* To be changed */
#define UPID_BEACON		0x3		/* Much better */

#undef INCLUDE_NULL
/*#define BROWSER		"lynx %s"*/
#define BROWSER		"netscape %s &"

/****************************** SENDER ******************************/

/* ---------------------------------------------------------------- */
/*
 * Send a beacon
 */
int
send_beacon(int		fd,
	    char *	string)
{
  unsigned char		buf[383];	/* Send buffer */
  int			actual;

  /* Copy the string */
  strncpy(buf, string, 380);
  buf[380] = '\0';

  printf("Ready to send on Ultra-PID %d\n", UPID_BEACON);

#ifdef INCLUDE_NULL
  /* Send a beacon including the trailing '\0' */
  actual = send(fd, &buf, strlen(buf) + 1, 0);
  printf("Sent IrDA Beacon ``%s'' (including \\0)\n", buf);
#else
  /* Send a beacon excluding the trailing '\0' */
  actual = send(fd, &buf, strlen(buf), 0);
  printf("Sent IrDA Beacon ``%s''\n", buf);
#endif

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Create XML string
 */
int
create_xml(char *	buf,
	   int		num,
	   char *	args[])
{
  /* Check number of args */
  switch(num)
    {
    case 0:
      strcpy(buf, "");
      break;

    case 1:
#ifdef OLD_XML
      strcpy(buf, args[0]);
#else OLD_XML
      sprintf(buf, "<a:link xlink:href=\"%s\"\\>", args[0]);
#endif OLD_XML
      break;

    case 2:
#ifdef OLD_XML
      sprintf(buf, "<squirt label=\"%s\" url=\"%s\"\\>", args[0], args[1]);
#else OLD_XML
      sprintf(buf, "<a:link xlink:title=\"%s\" xlink:href=\"%s\"\\>",
	      args[0], args[1]);
#endif OLD_XML
      break;	      

    default:
#ifdef OLD_XML
      sprintf(buf, "<squirt label=\"%s\" url=\"%s\" type=\"%s\"\\>",
	      args[0], args[1], args[2]);
#else OLD_XML
      sprintf(buf, "<a:link xlink:title=\"%s\" xlink:href=\"%s\" a:ContentType=\"%s\"\\>",
	      args[0], args[1], args[2]);
#endif OLD_XML
      break;
    }

  return 0;
}


/***************************** RECEIVER *****************************/

/* ---------------------------------------------------------------- */
/*
 * Parse an XML token
 */
char *
parse_token(char *	string,
	    char *	id,
	    char *	result)
{
  char *		pos;
  char *		pos2;
  char *		pos3;

  /* Find the ID */
  pos = strstr(string, id);
  if(pos == NULL)
    return(NULL);

  /* Find the first `"' */
  pos2 = strchr(pos, '"');
  if(pos2 == NULL)
    return(NULL);
  pos2++;

  /* Find the second `"' */
  pos3 = strchr(pos2, '"');
  if(pos3 == NULL)
    return(NULL);

  /* Copy over */
  memcpy(result, pos2, pos3 - pos2);
  result[pos3 - pos2] = '\0';
  return(result);
}

/* ---------------------------------------------------------------- */
/*
 * Parse an XML string
 */
int
parse_xml(char *	string,
	  char *	url)
{
  unsigned char		field[383];
  char *		token;

printf("String = [%s]\n", string);
  /* Check if XML or not */
  if(string[0] != '<')
    {
      printf("  The URL is ``%s''\n", string);
      strcpy(url, string);
    }
  else
    {
      /* Check the type of XML... */
      if(string[1] != 'a')
	{
	  /* Old XML format */
	  /* Find the label */
	  token = parse_token(string, "label=", field);
	  if(token != NULL)
	    printf("  The Label is ``%s''\n", token);

	  /* Find the label */
	  token = parse_token(string, "url=", field);
	  if(token != NULL)
	    {
	      printf("  The XML URL is ``%s''\n", token);
	      strcpy(url, token);
	    }

	  /* Find the label */
	  token = parse_token(string, "type=", field);
	  if(token != NULL)
	    printf("  The Type is ``%s''\n", token);
	}
      else
	{
	  /* New XML format */
	  /* Find the label */
	  token = parse_token(string, "xlink:title=", field);
	  if(token != NULL)
	    printf("  The Label is ``%s''\n", token);

	  /* Find the label */
	  token = parse_token(string, "xlink:href=", field);
	  if(token != NULL)
	    {
	      printf("  The XML URL is ``%s''\n", token);
	      strcpy(url, token);
	    }

	  /* Find the label */
	  token = parse_token(string, "a:ContentType=", field);
	  if(token != NULL)
	    printf("  The Type is ``%s''\n", token);
	}
    }
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Receive a beacon
 */
int
recv_beacon(int		fd)
{
  unsigned char		buf[383];	/* Receive buffer */
  int			actual;
  unsigned char		url[383];

  /* Receive packets */
  printf("Waiting for IrDA Beacons on UPID %d...\n", UPID_BEACON);
  while(1)
    {
      /* Block on read */
      actual = recv(fd, &buf, sizeof(buf) - 1, 0);
      buf[actual] = '\0';	/* Just to avoid crash on incorrect data */
      printf("->Getting beacon of size %d\n", actual);
      strcpy(url, "");
      parse_xml(buf, url);
#ifdef BROWSER
      sprintf(buf, BROWSER, url);
      system(buf);
#endif
    }
}

/******************************* MAIN *******************************/

/* ---------------------------------------------------------------- */
/*
 * main body of the program
 */
int
main(int	argc,
     char *	argv[])
{
  struct sockaddr_irda	self;		/* Socket address */
  char			string[512];	/* XML string */
  int			fd;		/* The socket itself */

  /* Create socket  - Ultra type */
  fd = socket(AF_IRDA, SOCK_DGRAM, 1);
  if(fd < 0)
    {
      perror("Creating socket");
      return(-1);
    }

  /* Bind to the specific Ultra PID */
  self.sir_family = AF_IRDA;
  self.sir_lsap_sel = UPID_BEACON;
  if(bind(fd, (struct sockaddr*) &self, sizeof(struct sockaddr_irda)))
    {
      perror("Connect");
      return(-1);
    }

  /* Look if there is any command line args */
  if(argc > 1)
    {
      create_xml(string, argc - 1, argv + 1);
      send_beacon(fd, string);
    }
  else
    recv_beacon(fd);

  /* Close the socket */
  close(fd);

  return(0);
}















