/*
 * Jean II '00
 *
 * Send beacon info or receive beacon info
 * Send : Open a Ultra socket and send one packet...
 * Recv : Open a Ultra socket and wait for packets...
 *
 * This version use Obex over Ultra.
 *
 * Note : Ultra can carry only 382 bytes / packet, and this including
 * the ultra header (the UPID). As we use a one byte UPID, that's 381 bytes.
 *
 * This program is released under the GPL license.
 */

#include <sys/socket.h>
#include <sys/types.h>

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include <linux/types.h>
#include <linux/irda.h>

#ifndef AF_IRDA
#define AF_IRDA 23
#endif /* AF_IRDA */

/*
 * This is the Ultra protocol ID we are currently using (7 bits)
 * Currently, IrDA-Ultra defines only UPID 0x1 for Obex.
 */
#define UPID_OBEX	0x1		/* Official */

/*#define BROWSER		"lynx %s"*/
#define BROWSER		"netscape %s &"

/****************************** SENDER ******************************/

/* ---------------------------------------------------------------- */
/*
 * Send a beacon
 *
 * IMPORTANT : this is just a test sample, and doesn't represent how
 * a beacon should be, because Obex is flexible.
 * For example, we might remove the type header (optional) and the headers
 * may be swapped around.
 * The string is either a plain URL or a XML description.
 */
int
send_beacon(int		fd,
	    char *	string)
{
  unsigned char		buf[383];	/* Send buffer */
  int			actual;
  int			length;
  int			total;

  /* Check the length and compute total length */
  length = strlen(string);
  if(length > (381 - 30))
    length = 381 - 30;
  total = length + 29;		/* 30 = Obex Headers - SAR */

  /* Start putting some OBEX headers */
  buf[0] = 0x00;		/* Ultra SAR - Segmentation and reassembly */
  buf[1] = 0x82;		/* Obex command = put+final */
  buf[2] = total >> 8;		/* Obex command length - high */
  buf[3] = total & 0xFF;	/* Obex command length - low */
  buf[4] = 0x41;		/* Obex target header type */
  buf[5] = 0x00;		/* Obex target header length - high */
  buf[6] = 0x0C;		/* Obex target header length - low */
  strcpy(buf + 7, "hp:beacon");	/* Obex target header payload */
  buf[16] = 0x42;		/* Obex type header type */
  buf[17] = 0x00;		/* Obex type header length - high */
  buf[18] = 0x0B;		/* Obex type header length - low */
  strcpy(buf + 19, "text/xml");	/* Obex type header payload */
  buf[27] = 0x49;		/* Obex end of body header type */
  buf[28] = (length +3) >> 8;	/* Obex end of body header length - high */
  buf[29] = (length +3) & 0xFF;	/* Obex end of body header length - low */
  memcpy(buf + 30, string, length);	/* Obex end of body header payload */

  printf("Ready to send Obex on Ultra-PID %d\n", UPID_OBEX);

  /* Send a beacon including the trailing '\0' */
  actual = send(fd, &buf, total + 1, 0);
  printf("Sent IrDA Beacon ``%s''\n", string);

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Send a e-squirt
 *
 * IMPORTANT : this is just a test sample, and doesn't represent how
 * a squirt should be, because Obex is flexible.
 * For example, we might remove the type header (optional) and the headers
 * may be swapped around.
 * The string is either a plain URL or a XML description.
 */
int
send_squirt(int			fd,
	    char *		string,
	    unsigned int	seq)
{
  unsigned char		buf[383];	/* Send buffer */
  int			actual;
  int			length;
  int			total;

  /* Check the length and compute total length */
  length = strlen(string);
  if(length > (381 - 36))
    length = 381 - 36;
  total = length + 35;		/* 36 = Obex Headers - SAR */

  /* Start putting some OBEX headers */
  buf[0] = 0x00;		/* Ultra SAR - Segmentation and reassembly */
  buf[1] = 0x82;		/* Obex command = put+final */
  buf[2] = total >> 8;		/* Obex command length - high */
  buf[3] = total & 0xFF;	/* Obex command length - low */
  buf[4] = 0x41;		/* Obex target header type */
  buf[5] = 0x00;		/* Obex target header length - high */
  buf[6] = 0x0D;		/* Obex target header length - low */
  strcpy(buf + 7, "hp:esquirt");	/* Obex target header payload */
  buf[17] = 0xCC;		/* Obex app params header type */
  buf[18] = (seq >> 24) & 0xFF;	/* Obex app params header payload */
  buf[19] = (seq >> 16) & 0xFF;	/* Obex app params header payload */
  buf[20] = (seq >> 8) & 0xFF;	/* Obex app params header payload */
  buf[21] = seq & 0xFF;		/* Obex app params header payload */
  buf[22] = 0x42;		/* Obex type header type */
  buf[23] = 0x00;		/* Obex type header length - high */
  buf[24] = 0x0B;		/* Obex type header length - low */
  strcpy(buf + 25, "text/xml");	/* Obex type header payload */
  buf[33] = 0x49;		/* Obex end of body header type */
  buf[34] = (length +3) >> 8;	/* Obex end of body header length - high */
  buf[35] = (length +3) & 0xFF;	/* Obex end of body header length - low */
  memcpy(buf + 36, string, length);	/* Obex end of body header payload */

  printf("Ready to send Obex on Ultra-PID %d\n", UPID_OBEX);

  /* Send a beacon including the trailing '\0' */
  actual = send(fd, &buf, total + 1, 0);
  printf("Sent IrDA Beacon ``%s''\n", string);

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Create XML string
 */
int
create_xml(char *	buf,
	   int		num,
	   char *	args[])
{
  /* Check number of args */
  switch(num)
    {
    case 0:
      strcpy(buf, "");
      break;

    case 1:
      strcpy(buf, args[0]);
      break;

    case 2:
      sprintf(buf, "<squirt label=\"%s\" url=\"%s\"\\>", args[0], args[1]);
      break;	      

    default:
      sprintf(buf, "<squirt label=\"%s\" url=\"%s\" type=\"%s\"\\>",
	      args[0], args[1], args[2]);
      break;
    }

  return 0;
}


/***************************** RECEIVER *****************************/

/* ---------------------------------------------------------------- */
/*
 * Parse an Obex header
 */
int
parse_header(char *	buf,
	     int	i,
	     int	actual,
	     char *	pld)
{
  int			len;

  /* Get and check length */
  len = (buf[i + 1] << 8) | (buf[i + 2]);
  if((i + len) > actual)
    {
      printf("Invalid Obex header length (%d)\n", len);
      return -1;
    }
  /* Get and display payload */
  i += 3; len -= 3;
  memcpy(pld, buf + i, len);
  pld[len] = '\0';

  /* Done */
  return(i + len);
}

/* ---------------------------------------------------------------- */
/*
 * Parse an XML token
 */
char *
parse_token(char *	string,
	    char *	id,
	    char *	result)
{
  char *		pos;
  char *		pos2;
  char *		pos3;

  /* Find the ID */
  pos = strstr(string, id);
  if(pos == NULL)
    return(NULL);

  /* Find the first `"' */
  pos2 = strchr(pos, '"');
  if(pos2 == NULL)
    return(NULL);
  pos2++;

  /* Find the second `"' */
  pos3 = strchr(pos2, '"');
  if(pos3 == NULL)
    return(NULL);

  /* Copy over */
  memcpy(result, pos2, pos3 - pos2);
  result[pos3 - pos2] = '\0';
  return(result);
}

/* ---------------------------------------------------------------- */
/*
 * Parse an XML string
 */
int
parse_xml(char *	string,
	  char *	url)
{
  unsigned char		field[383];
  char *		token;

  /* Check if XML or not */
  if(string[0] != '<')
    {
      printf("  The URL is ``%s''\n", string);
      strcpy(url, string);
    }
  else
    {
      /* Find the label */
      token = parse_token(string, "label=", field);
      if(token != NULL)
	printf("  The Label is ``%s''\n", token);

      /* Find the label */
      token = parse_token(string, "url=", field);
      if(token != NULL)
	{
	  printf("  The XML URL is ``%s''\n", token);
	  strcpy(url, token);
	}

      /* Find the label */
      token = parse_token(string, "type=", field);
      if(token != NULL)
	printf("  The Type is ``%s''\n", token);
    }
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Receive a beacon
 */
int
recv_beacon(unsigned char *	buf,
	    int			actual,
	    char *		url)
{
  unsigned char		pld[383];
  int			len;
  unsigned int		seq;
  int			i = 4;

  /* Init */
  strcpy(url, "");

  /* Check the initial header */
  if(buf[0] != 0x00)
    printf("Invalid SAR\n");
  if(buf[1] != 0x82)
    printf("Invalid Obex command\n");
  len = (buf[2] << 8) | (buf[3]);
  if(len != (actual - 1))
    printf("Invalid Obex command length\n");

  /* Get all Obex headers */
  while(i < actual)
    {
      switch(buf[i])
	{
	case 0x41:		/* Target header */
	  i = parse_header(buf, i, actual, pld);
	  if(i < 0)
	    return -1;
	  printf("Found target header : ``%s''\n", pld);
	  break;
	case 0x42:		/* Type header */
	  i = parse_header(buf, i, actual, pld);
	  if(i < 0)
	    return -1;
	  printf("Found type header : ``%s''\n", pld);
	  break;
	case 0xCC:		/* Apps params header */
	  seq =  ((buf[i + 1] << 24) | (buf[i + 2] << 16) | 
		  (buf[i + 3] << 8) | (buf[i + 4]));
	  i += 5;
	  printf("Found app params : 0x%08X - %d\n", seq, seq);
	  break;
	case 0x49:		/* Body header */
	  i = parse_header(buf, i, actual, pld);
	  if(i < 0)
	    return -1;
	  parse_xml(pld, url);
	  break;
	default:
	  printf("Unknown header 0x%02X\n", buf[i]);
	  return -1;
	}
    }
  return 0;
}

/* ---------------------------------------------------------------- */
/*
 * Receive a beacon
 */
int
recv_loop(int		fd)
{
  unsigned char		buf[383];	/* Receive buffer */
  int			actual;
  unsigned char		url[383];

  /* Receive packets */
  printf("Waiting for Obex Beacons on Ultra-PID %d...\n", UPID_OBEX);
  while(1)
    {
      /* Block on read */
      actual = recv(fd, &buf, sizeof(buf) - 1, 0);
      printf("->Getting beacon of size %d\n", actual);
      recv_beacon(buf, actual, url);
#ifdef BROWSER
      sprintf(buf, BROWSER, url);
      system(buf);
#endif
    }
}

/******************************* MAIN *******************************/

/* ---------------------------------------------------------------- */
/*
 * main body of the program
 */
int
main(int	argc,
     char *	argv[])
{
  struct sockaddr_irda	self;		/* Socket address */
  int			fd;		/* The socket itself */
  char			string[512];	/* XML string */
  unsigned int		seq;

  /* Create socket  - Ultra type */
  fd = socket(AF_IRDA, SOCK_DGRAM, 1);
  if(fd < 0)
    {
      perror("Creating socket");
      return(-1);
    }

  /* Bind to the specific Ultra PID */
  self.sir_family = AF_IRDA;
  self.sir_lsap_sel = UPID_OBEX;
  if(bind(fd, (struct sockaddr*) &self, sizeof(struct sockaddr_irda)))
    {
      perror("Connect");
      return(-1);
    }

  /* Look if there is any command line args */
  if(argc > 1)
    {
      /* Check is there is a sequence number */
      if(isdigit(argv[1][0]))
	{
	  sscanf(argv[1], "%d", &seq);
	  create_xml(string, argc - 2, argv + 2);
	  send_squirt(fd, string, seq);
	}
      else
	{
	  create_xml(string, argc - 1, argv + 1);
	  send_beacon(fd, string);
	}
    }
  else
    recv_loop(fd);

  /* Close the socket */
  close(fd);

  return(0);
}















